@@ -14,7 +14,32 @@
 # You should have received a copy of the GNU General Public License
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
-from __future__ import print_function
+"""
+pprofile - Line-granularity, thread-aware deterministic and statistic
+pure-python profiler
+
+Usage as a command line:
+$ pprofile --exclude-syspath some_python_executable arg1 ...
+$ pprofile --exclude-syspath -m some_python_module -- arg1 ...
+$ python -m pprofile --exclude-syspath some_python_executable arg1 ...
+$ python -m pprofile -m some_python_module -- arg1 ...
+See --help for all options.
+
+Usage as a python module:
+
+Deterministic profiling:
+>>> prof = pprofile.Profile()
+>>> with prof():
+>>>     # Code to profile
+>>> prof.print_stats()
+
+Statistic profiling:
+>>> prof = StatisticalProfile()
+>>> with prof():
+>>>     # Code to profile
+>>> prof.print_stats()
+"""
+
 from collections import defaultdict, deque
 from functools import partial, wraps
 # Note: use time, not clock.
@@ -29,13 +54,26 @@ from functools import partial, wraps
 from time import time
 from warnings import warn
 import argparse
+import io
 import inspect
 import linecache
 import os
 import re
+import runpy
+import shlex
 import sys
 import threading
 import zipfile
+try:
+    from IPython.core.magic import register_line_cell_magic
+except ImportError:
+    register_line_cell_magic = lambda x: x
+
+__all__ = (
+    'ProfileBase', 'ProfileRunnerBase', 'Profile', 'ThreadProfile',
+    'StatisticProfile', 'StatisticThread', 'run', 'runctx', 'runfile',
+    'runpath',
+)
 
 if sys.version_info < (3, ):
     # Python 2.x linecache returns non-decoded strings, which cause errors when
@@ -57,13 +95,20 @@ else:
         """
         if encoding is None:
             encoding = stream.encoding
-        # XXX: Python3 < 3.2 does not have stream.buffer so this will raise.
+        # XXX: Python3 < 3.2 and ipykernel.iostream.OutStream at least up to
+        # 4.5.0 do not have stream.buffer.
         # I do not see a way to change errors without also potentially changing
         # the encoding, and there does not seem to be a way to change encoding
         # without having to access the binary stream.
-        # Also, I do not expect many 3.0 and 3.1 to be still used. Feel free to
-        # report a bug if it raises.
-        return codecs.getwriter(encoding)(stream.buffer, errors=errors)
+        try:
+            buf = stream.buffer
+        except AttributeError:
+            warn(
+                'Cannot access "%r.buffer", invalid entities from source '
+                'files will cause errors when annotating.' % (stream, )
+            )
+            return stream
+        return codecs.getwriter(encoding)(buf, errors=errors)
 
 def _getFuncOrFile(func, module, line):
     if func == '<module>' or func is None:
@@ -71,6 +116,9 @@ def _getFuncOrFile(func, module, line):
     else:
         return '%s:%s' % (func, line)
 
+def _isCallgrindName(filepath):
+    return os.path.basename(filepath).startswith('cachegrind.out.')
+
 class _FileTiming(object):
     """
     Accumulation of profiling statistics (line and call durations) for a given
@@ -153,7 +201,7 @@ class _FileTiming(object):
     def getCallListByLine(self):
         result = defaultdict(list)
         for (line, name, callee), (code, hit, duration) in \
-                self.call_dict.iteritems():
+                self.call_dict.items():
             result[line].append((
                 code.co_name, code.co_firstlineno,
                 hit, duration,
@@ -162,15 +210,15 @@ class _FileTiming(object):
         return result
 
     def getTotalTime(self):
-        return sum(x[2] for x in self.line_dict.itervalues())
+        return sum(x[2] for x in self.line_dict.values())
 
     def getTotalHitCount(self):
-        return sum(x[1] for x in self.line_dict.itervalues())
+        return sum(x[1] for x in self.line_dict.values())
 
     def getSortKey(self):
         # total duration first, then total hit count for statistical profiling
         result = [0, 0]
-        for _, hit, duration in self.line_dict.itervalues():
+        for _, hit, duration in self.line_dict.values():
             result[0] += duration
             result[1] += hit
         return result
@@ -255,7 +303,7 @@ class ProfileBase(object):
 
     def _getFileTiming(self, frame):
         try:
-            return self.global_dict[id(frame.f_globals)]
+            return self.global_dict[(id(frame.f_globals), frame.f_code.co_filename)]
         except KeyError:
             f_globals = frame.f_globals
             name = self._getFilename(frame.f_code.co_filename, f_globals)
@@ -300,7 +348,7 @@ class ProfileBase(object):
     def _getFileNameList(self, filename, may_sort=True):
         if filename is None:
             filename = self.getFilenameSet()
-        elif isinstance(filename, basestring):
+        elif isinstance(filename, str):
             return [filename]
         if may_sort:
             try:
@@ -368,11 +416,11 @@ class ProfileBase(object):
             files which may not match with profiled code.
         """
         print('version: 1', file=out)
-        if commandline is not None:
-            print('cmd:', commandline, file=out)
         print('creator: pprofile', file=out)
         print('event: usphit :us/hit', file=out)
         print('events: hits us usphit', file=out)
+        if commandline is not None:
+            print('cmd:', commandline, file=out)
         file_dict = self.file_dict
         if relative_path:
             convertPath = _relpath
@@ -509,8 +557,12 @@ class ProfileBase(object):
         """
         Similar to profile.Profile.dump_stats - but different output format !
         """
-        with _open(filename, 'w', errors='replace') as out:
-            self.annotate(out)
+        if _isCallgrindName(filename):
+            with open(filename, 'w') as out:
+                self.callgrind(out)
+        else:
+            with _open(filename, 'w', errors='replace') as out:
+                self.annotate(out)
 
     def print_stats(self):
         """
@@ -520,31 +572,40 @@ class ProfileBase(object):
         self.annotate(_reopen(sys.stdout, errors='replace'))
 
 class ProfileRunnerBase(object):
+    def __call__(self):
+        return self
+
+    def __enter__(self):
+        raise NotImplementedError
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        raise NotImplementedError
+
     # profile/cProfile-like API
     def runctx(self, cmd, globals, locals):
         """Similar to profile.Profile.runctx ."""
-        with self:
+        with self():
             exec(cmd, globals, locals)
         return self
 
     def runcall(self, func, *args, **kw):
         """Similar to profile.Profile.runcall ."""
-        with self:
+        with self():
             return func(*args, **kw)
 
     def runfile(self, fd, argv, fd_name='<unknown>', compile_flags=0,
-            dont_inherit=1):
+            dont_inherit=1, globals={}):
         with fd:
             code = compile(fd.read(), fd_name, 'exec', flags=compile_flags,
                 dont_inherit=dont_inherit)
         original_sys_argv = list(sys.argv)
+        ctx_globals = globals.copy()
+        ctx_globals['__file__'] = fd_name
+        ctx_globals['__name__'] = '__main__'
+        ctx_globals['__package__'] = None
         try:
             sys.argv[:] = argv
-            return self.runctx(code, {
-                '__file__': fd_name,
-                '__name__': '__main__',
-                '__package__': None,
-            }, None)
+            return self.runctx(code, ctx_globals, None)
         finally:
             sys.argv[:] = original_sys_argv
 
@@ -556,6 +617,19 @@ class ProfileRunnerBase(object):
         finally:
             sys.path[:] = original_sys_path
 
+    def runmodule(self, module, argv):
+        original_sys_argv = list(sys.argv)
+        original_sys_path0 = sys.path[0]
+        try:
+            sys.path[0] = os.getcwd()
+            sys.argv[:] = argv
+            with self():
+                runpy.run_module(module, run_name='__main__', alter_sys=True)
+        finally:
+            sys.argv[:] = original_sys_argv
+            sys.path[0] = original_sys_path0
+        return self
+
 class Profile(ProfileBase, ProfileRunnerBase):
     """
     Deterministic, recursive, line-granularity, profiling class.
@@ -620,9 +694,17 @@ class Profile(ProfileBase, ProfileRunnerBase):
         else:
             warn('Duplicate "disable" call')
 
-    __enter__ = enable
+    def __enter__(self):
+        """
+        __enter__() -> self
+        """
+        self.enable()
+        return self
 
     def __exit__(self, exc_type, exc_val, exc_tb):
+        """
+        __exit__(*excinfo) -> None. Disables profiling.
+        """
         self.disable()
 
     def _traceEvent(self, frame, event):
@@ -716,16 +798,16 @@ class ThreadProfile(Profile):
         threading.settrace(None)
         self._local_trace = None
 
-class StatisticalProfile(ProfileBase):
+class StatisticProfile(ProfileBase, ProfileRunnerBase):
     """
-    Statistical profiling class.
+    Statistic profiling class.
 
     This class does not gather its own samples by itself.
     Instead, it must be provided with call stacks (as returned by
     sys._getframe() or sys._current_frames()).
     """
     def __init__(self):
-        super(StatisticalProfile, self).__init__()
+        super(StatisticProfile, self).__init__()
         self.total_time = 1
 
     def sample(self, frame):
@@ -745,11 +827,32 @@ class StatisticalProfile(ProfileBase):
             frame = caller
             called_code = caller_code
 
-class StatisticalThread(threading.Thread, ProfileRunnerBase):
+    def __call__(self, period=.001, single=True, group=None, name=None):
+        """
+        Instanciate StatisticThread.
+
+        >>> s_profile = StatisticProfile()
+        >>> with s_profile(single=False):
+        >>>    # Code to profile
+        Is equivalent to:
+        >>> s_profile = StatisticProfile()
+        >>> s_thread = StatisticThread(profiler=s_profile, single=False)
+        >>> with s_thread:
+        >>>    # Code to profile
+        """
+        return StatisticThread(
+            profiler=self, period=period, single=single, group=group,
+            name=name,
+        )
+
+# BBB
+StatisticalProfile = StatisticProfile
+
+class StatisticThread(threading.Thread, ProfileRunnerBase):
     """
     Usage in a nutshell:
-      profiler = StatisticalProfile()
-      pt = StatisticalThread(profiler)
+      profiler = StatisticProfile()
+      pt = StatisticThread(profiler)
       with pt:
         # do stuff
       profiler.print_stats()
@@ -766,6 +869,7 @@ class StatisticalThread(threading.Thread, ProfileRunnerBase):
           become meaningful.
           The larger, the less profiling overhead, but requires long profiling
           session to get meaningful results.
+          Available on instances as the "profiler" read-only property.
         single (bool)
           Profile only the thread which created this instance.
         group, name
@@ -773,20 +877,24 @@ class StatisticalThread(threading.Thread, ProfileRunnerBase):
         """
         if single:
             self._test = lambda x, ident=threading.current_thread().ident: ident == x
-        super(StatisticalThread, self).__init__(
+        super(StatisticThread, self).__init__(
             group=group,
             name=name,
         )
         self._stop_event = threading.Event()
         self._period = period
-        self.profiler = profiler
+        self._profiler = profiler
         profiler.total_time = 0
         self.daemon = True
 
+    @property
+    def profiler(self):
+        return self._profiler
+
     def start(self):
         self._start_time = time()
         self._can_run = True
-        super(StatisticalThread, self).start()
+        super(StatisticThread, self).start()
 
     def stop(self):
         """
@@ -796,12 +904,20 @@ class StatisticalThread(threading.Thread, ProfileRunnerBase):
         if self.is_alive():
             self._can_run = False
             self._stop_event.set()
-            self.profiler.total_time += time() - self._start_time
+            self._profiler.total_time += time() - self._start_time
             self._start_time = None
 
-    __enter__ = start
+    def __enter__(self):
+        """
+        __enter__() -> self
+        """
+        self.start()
+        return self
 
     def __exit__(self, exc_type, exc_val, exc_tb):
+        """
+        __exit__(*excinfo) -> None. Stops and joins profiling thread.
+        """
         self.stop()
         self.join()
 
@@ -810,11 +926,11 @@ class StatisticalThread(threading.Thread, ProfileRunnerBase):
         test = self._test
         if test is None:
             test = lambda x, ident=self.ident: ident != x
-        sample = self.profiler.sample
+        sample = self._profiler.sample
         stop_event = self._stop_event
         wait = partial(stop_event.wait, self._period)
         while self._can_run:
-            for ident, frame in current_frames().iteritems():
+            for ident, frame in current_frames().items():
                 if test(ident):
                     sample(frame)
             frame = None
@@ -824,23 +940,26 @@ class StatisticalThread(threading.Thread, ProfileRunnerBase):
 
     def callgrind(self, *args, **kw):
         warn('deprecated', DeprecationWarning)
-        return self.profiler.callgrind(*args, **kw)
+        return self._profiler.callgrind(*args, **kw)
 
     def annotate(self, *args, **kw):
         warn('deprecated', DeprecationWarning)
-        return self.profiler.annotate(*args, **kw)
+        return self._profiler.annotate(*args, **kw)
 
     def dump_stats(self, *args, **kw):
         warn('deprecated', DeprecationWarning)
-        return self.profiler.dump_stats(*args, **kw)
+        return self._profiler.dump_stats(*args, **kw)
 
     def print_stats(self, *args, **kw):
         warn('deprecated', DeprecationWarning)
-        return self.profiler.print_stats(*args, **kw)
+        return self._profiler.print_stats(*args, **kw)
 
     def iterSource(self, *args, **kw):
         warn('deprecated', DeprecationWarning)
-        return self.profiler.iterSource(*args, **kw)
+        return self._profiler.iterSource(*args, **kw)
+
+# BBB
+StatisticalThread = StatisticThread
 
 # profile/cProfile-like API (no sort parameter !)
 def _run(threads, verbose, func_name, filename, *args, **kw):
@@ -892,15 +1011,15 @@ def _relpath(name):
     """
     return os.path.normpath(os.path.splitdrive(name)[1]).lstrip(_allsep)
 
-def main():
+def _main(argv, stdin=None):
     format_dict = {
         'text': 'annotate',
         'callgrind': 'callgrind',
     }
 
-    parser = argparse.ArgumentParser()
+    parser = argparse.ArgumentParser(argv[0])
     parser.add_argument('script', help='Python script to execute (optionaly '
-        'followed by its arguments)', nargs=1)
+        'followed by its arguments)', nargs='?')
     parser.add_argument('argv', nargs=argparse.REMAINDER)
     parser.add_argument('-o', '--out', default='-',
         help='Write annotated sources to this file. Defaults to stdout.')
@@ -917,6 +1036,10 @@ def main():
     parser.add_argument('-s', '--statistic', default=0, type=float,
         help='Use this period for statistic profiling, or use deterministic '
         'profiling when 0.')
+    parser.add_argument('-m', dest='module',
+        help='Searches sys.path for the named module and runs the '
+        'corresponding .py file as a script. When given, positional arguments '
+        'become sys.argv[1:]')
 
     group = parser.add_argument_group(
         title='Filtering',
@@ -936,7 +1059,7 @@ def main():
         help='Include files whose name would have otherwise excluded. '
         'If no exclusion was specified, all paths are excluded first.')
 
-    options = parser.parse_args()
+    options = parser.parse_args(argv[1:])
     if options.exclude_syspath:
         options.exclude.extend('^' + re.escape(x) for x in sys.path)
     if options.include and not options.exclude:
@@ -951,9 +1074,41 @@ def main():
                 for regex in options.include:
                     print('\t' + regex, file=sys.stderr)
 
-    args = options.script + options.argv
+    if options.module is None:
+        if options.script is None:
+            parser.error('too few arguments')
+        args = [options.script] + options.argv
+        runner_method_kw = {
+            'path': args[0],
+            'argv': args,
+        }
+        runner_method_id = 'runpath'
+    elif stdin is not None and options.module == '-':
+        # Undocumented way of using -m, used internaly by %%pprofile
+        args = ['<stdin>']
+        if options.script is not None:
+            args.append(options.script)
+        args.extend(options.argv)
+        import __main__
+        runner_method_kw = {
+            'fd': stdin,
+            'argv': args,
+            'fd_name': '<stdin>',
+            'globals': __main__.__dict__,
+        }
+        runner_method_id = 'runfile'
+    else:
+        args = [options.module]
+        if options.script is not None:
+            args.append(options.script)
+        args.extend(options.argv)
+        runner_method_kw = {
+            'module': options.module,
+            'argv': args,
+        }
+        runner_method_id = 'runmodule'
     if options.format is None:
-        if os.path.basename(options.out).startswith('cachegrind.out.'):
+        if _isCallgrindName(options.out):
             options.format = 'callgrind'
         else:
             options.format = 'text'
@@ -972,7 +1127,7 @@ def main():
             klass = Profile
         prof = runner = klass(verbose=options.verbose)
     try:
-        runner.runpath(args[0], args)
+        getattr(runner, runner_method_id)(**runner_method_kw)
     finally:
         if options.out == '-':
             out = _reopen(sys.stdout, errors='replace')
@@ -1024,5 +1179,24 @@ def main():
         # do not change exit status.
         sys.exit(1)
 
+def pprofile(line, cell=None):
+    """
+    Profile line execution.
+    """
+    if cell is None:
+        # TODO: detect and use arguments (statistical profiling, ...) ?
+        return run(line)
+    else:
+        return _main(['%%pprofile', '-m', '-'] + shlex.split(line), io.StringIO(cell))
+try:
+    register_line_cell_magic(pprofile)
+except Exception:
+    # ipython can be imported, but may not be currently running.
+    pass
+del pprofile
+
+def main():
+    _main(sys.argv)
+
 if __name__ == '__main__':
     main()
